import os
import tempfile
import slugify
import logging
import clisync

from typing import Any, List, Optional, Union, Literal, Tuple

from pyhectiqlab import API_URL
from pyhectiqlab import Config
from pyhectiqlab.project import Project
from pyhectiqlab.artifact import Artifact
from pyhectiqlab.block import Block
from pyhectiqlab.tag import Tag
from pyhectiqlab.model import Model
from pyhectiqlab.dataset import Dataset
from pyhectiqlab.versions import PackageVersion
from pyhectiqlab.step import Step

from pyhectiqlab.client import Client
from pyhectiqlab.decorators import classproperty, online_method, functional_alias, no_git_diff
from pyhectiqlab.metrics import MetricsManager

logger = logging.getLogger()


class Run:
    _metrics: Optional[MetricsManager] = None
    _tmp_artifacts: Optional[str] = None
    _id: Optional[str] = None
    _rank: Optional[int] = None
    _step: Optional[int] = None
    _repos: Optional[List[str]] = None
    _config: Optional[dict] = None
    _client: Client = Client

    def __init__(
        self,
        rank: Optional[int] = None,
        title: Optional[str] = None,
        project: Optional[str] = None,
        block: Optional[str] = None,
        category: Literal[None, "develop", "evaluation", "test", "training"] = None,
        config: Optional[dict] = None,
    ):
        if rank is not None and isinstance(rank, int):
            self.retrieve(rank, project)
        elif isinstance(rank, str) or rank is None:
            title = title or rank  # support for positional argument
            self.create(title=title, project=project, block=block, category=category, config=config)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._metrics:
            self._metrics.flush_cache()
        if Run.id is None:
            return
        if exc_type is not None:
            if Step.id:
                Step.end(status="failed")
            Run._update(run_id=Run.id, status="failed", package_versions=PackageVersion.all(repos=Run._repos))
        else:
            if Step.id:
                Step.end(status="completed")
            Run._update(run_id=Run.id, status="completed", package_versions=PackageVersion.all(repos=Run._repos))

    @staticmethod
    @functional_alias("track_version")
    @clisync.include()
    def track_version(
        repos: Union[str, List[str], None] = None, run_id: Optional[str] = None, project: Optional[str] = None
    ):
        """Track the git status of the given repositories.

        Args:
            repos (List[str]): List of modules for which the files are located in a git repertoire.
            run_id (str): Id of the run to update. If None, the current run is used. Default: None.
        """
        if not isinstance(repos, list) and (repos is not None):
            repos = [repos]
        Run._repos = repos
        if info := PackageVersion.all(repos=repos):
            Run._update(run_id=run_id, package_versions=info)

    @staticmethod
    def setup(**kwargs):
        """Setup the run with the given `id` and `rank`."""
        Run._id = kwargs.get("id")
        Run._rank = kwargs.get("rank")

    @classproperty
    def tmp_artifacts(cls):
        if cls._tmp_artifacts is None:
            cls._tmp_artifacts = tempfile.mkdtemp()
        return cls._tmp_artifacts

    @classproperty
    def rank(cls) -> Optional[int]:
        return cls._rank

    @classproperty
    def id(cls) -> Optional[str]:
        return cls._id

    @classproperty
    def category(cls):
        return cls.get(["category"])["category"]

    @classproperty
    def config(cls):
        return cls._config

    @staticmethod
    @functional_alias("get_run_id")
    def get_id(id: Optional[str] = None):
        return id or Run.id

    @staticmethod
    @functional_alias("run_is_ready")
    def ready():
        """Check if the run is ready."""
        return Run.id is not None

    @staticmethod
    @functional_alias("create_run")
    @no_git_diff()
    @clisync.include()
    def create(
        title: Optional[str] = None,
        project: Optional[str] = None,
        block: Optional[str] = None,
        category: Literal[None, "develop", "evaluation", "test", "training"] = None,
        config: Union[dict, None] = None,
        tags: Optional[List[str]] = None,
        online: bool = False,
    ) -> dict:
        """
        Creates a new run.

        Args:
            title (str): The title of the run.
            project (str): The project of the run. If None, the current project is used.
            block (str): The block of the run. If None, the current block is used.
            category (str): The category of the run in [None, 'develop', 'evaluation', 'test', 'training']. Default: None.
            config (dict): The configuration dict of the run. Default: None.
            tags (list): The tags of the run. Default: None.
            online (bool): Set to True to create the run online. Default: True.
        """
        if online:
            Run._client.online(online)
            return False

        project = Project.get(project)
        if project is None:
            logging.error("No project has been set. Please set a project first. Continuing in offline mode.")
            Run._client.online(False)
            return
        Project.set(project)

        if block is not None:
            if not "::" in block:
                block = Block.format_id(block, project)
            Block.set(block)

        data = dict(
            title=title,
            project=project,
            block=block,
            category=category,
            config=config or {},
            tags=tags or [],
        )

        run = Run._client.post("/app/runs", json=data, wait_response=True)
        if run is None:
            logging.error("Run failed to create.")
            return False
        Run.setup(**run)
        Run._update(run_id=Run.id, status="initiated", package_versions=PackageVersion.all(repos=Run._repos))
        return run

    @staticmethod
    @functional_alias("retrieve_run")
    @no_git_diff()
    @clisync.include()
    def retrieve(
        rank: int,
        project: Optional[str] = None,
        fields: Optional[List[str]] = None,
        setup_with_result: Optional[bool] = True,
    ):
        """
        Retrieves a run from the server.
        Use this method to attach to an existing run.

        Args:
            rank (int): The rank of the run.
            project (str): The project of the run. If None, the project is taken from the context. Default: None.
            fields (List[str]): The fields to retrieve. Default: None.
        """

        project = Project.get(project)
        if project is None:
            logging.error("No project has been set. Please set a project first. Continuing in offline mode.")
            Run._client.online(False)
            return
        Project.set(project)

        run = Run._client.get(
            "/app/runs/retrieve", params=dict(rank=rank, project=project, fields=fields), wait_response=True
        )
        if run is None:
            logging.error(f"Run {rank} not found in project {project}")
            return
        if setup_with_result:
            Run.setup(**run)
        return run

    @staticmethod
    @functional_alias("retrieve_run_by_id")
    @no_git_diff()
    @clisync.include()
    def retrieve_by_id(
        run_id: Optional[str] = None,
        fields: Optional[List[str]] = None,
        setup_with_result: Optional[bool] = False,
    ):
        """
        Retrieves a run from the server.
        Use this method to attach to an existing run.

        Args:
            rank (int): The rank of the run.
            project (str): The project of the run. If None, the project is taken from the context. Default: None.
            fields (List[str]): The fields to retrieve. Default: None.
            setup_with_result (bool): Set to True to setup the run with the result. Default: False.
        """

        run_id = run_id or Run.id
        if not run_id:
            logging.error("No run to retrieve. ")
            return
        run_id = Run._client.get(f"/app/runs/{run_id}", params=dict(fields=fields), wait_response=True)
        if not run_id:
            logging.error(f"Run {run_id} not found. Verify the run ID and make sure you have access to it.")
            return
        if setup_with_result:
            Run.setup(**run_id)
        return run_id

    @staticmethod
    @online_method
    @functional_alias("retrieve_config_by_id")
    @clisync.include()
    def retrieve_config_by_id(run_id: Optional[str] = None) -> dict:
        """Retrieve a config from a run.
        If the run does not have a config, an empty config is returned.

        Args:
            run (str, optional): Run ID. If None, the current run is used. Default: None.
            project (str, optional): Project name. If None, the current project is used. Default: None.
        """
        run = run_id or Run.id
        run = Run.retrieve_by_id(run=run, fields=["config"], setup_with_result=False)
        if run is None:
            return {}

        return run.get("config", {})

    @staticmethod
    @online_method
    @functional_alias("retrieve_config")
    @clisync.include()
    def retrieve_config(rank: Optional[int] = None, project: Optional[str] = None) -> dict:
        """Retrieve a config from a run.
        If the run does not have a config, an empty config is returned.

        Args:
            rank (int): Run rank. If None, the current run is used. Default: None.
            project (str, optional): Project name. If None, the current project is used. Default: None.
        """
        rank = rank or Run.rank
        run = Run.retrieve(rank=rank, project=project, fields=["config"], setup_with_result=False)
        if run is None:
            return {}

        return run.get("config", {})

    @staticmethod
    @functional_alias("run_exists")
    @clisync.include()
    def exists(rank: int, project: Optional[str] = None) -> bool:
        """
        Check if a run exists.

        Args:
            rank (int): The rank of the run.
            project (str): The project of the run. If None, the project is taken from the context. Default: None.
        """
        res = Run._client.get("/app/runs/retrieve", params=dict(rank=rank, project=project), wait_response=True)
        return res is not None

    @staticmethod
    @online_method
    def _update(run_id: Optional[str] = None, wait_response: bool = False, **kwargs):
        run_id = run_id or Run.id
        if not (run_id):
            logging.error("No run to update. Please create or retrieve a run first.")
            return
        return Run._client.patch(f"/app/runs/{run_id}", json=kwargs, wait_response=wait_response)

    @staticmethod
    @online_method
    @functional_alias("set_category")
    @clisync.include()
    def set_category(category: Literal["develop", "evaluation", "test", "training"], run_id: Optional[str] = None):
        """Set the category of the run."""
        Run._update(run_id=run_id, category=category)

    @staticmethod
    @online_method
    @functional_alias("set_block")
    @clisync.include()
    def set_block(block: str, run_id: Optional[str] = None):
        """Set the block of the run."""
        Run._update(run_id=run_id, block=block)
        Block.set(block)

    @staticmethod
    @online_method
    @functional_alias("rename_run")
    @clisync.include()
    def rename(title: str, run_id: Optional[str] = None):
        """Rename the run with a new title.

        Args:
            title (str): The new title of the run.
        """
        Run._update(run_id=run_id, title=title)

    @staticmethod
    @online_method
    @functional_alias("add_artifact")
    @clisync.include()
    def add_artifact(
        path: str,
        name: Optional[str] = None,
        step: Optional[int] = None,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Log a file as an artifacts. If a file already exists with the name, it will be overwritten.

        Args:
            path (str) : Path to the file.
            name (str, optional): Name of the artifact or the group of artifacts. If None, the name is the basename of the file. Default: None.
            step (int, optional): The optional step stamp of the artifacts. If None, the artifacts is not considered as a step artifact. Default: None.
            run_id (str, optional): ID of the run. If None, the current run ID is used. Default: None.
            project (str, optional): The project of the artifact. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to upload sync. If False, the upload is made in background. Default: False.
        """
        if Run.get_id(run_id) is None or Project.get(project) is None:
            return
        return Artifact.create(
            path,
            Run.get_id(run_id),
            Project.get(project),
            name=name,
            group=name,
            step=step,
            wait_response=wait_response,
        )

    @staticmethod
    @online_method
    @functional_alias(alias="add_figure")
    def add_figure(
        figure,
        name: str,
        step: Optional[int] = None,
        dpi: int = 200,
        extension: str = "png",
        wait_response: bool = False,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        **kwargs,
    ):
        """
        Save a matplotlib figure on a temp dir and push the image to the lab.

        Args:
            figure (matplotlib.pyplot.Figure): The figure to save.
            name (str): The name of the artifact.
            step (int): The optional step stamp of the artifacts. If None, the artifacts is not considered as a step artifact. Default: None.
            dpi (int): The dpi of the figure. Default: 200.
            extension (str): The extension of the figure. Default: "png".
            wait_response [bool]: Set to true to upload sync. If False, the upload is made in background. Default: False.
            run_id (str): The ID of the run. If None, the current run ID is used. Default: None.
            project (str): The project of the artifact. If None, the current project is used. Default: None.
        """

        basename, name_ext = os.path.splitext(name)
        if name_ext != "":
            extension = name_ext.strip(".")
        else:
            extension = (extension or "png").strip(".")

        slug_name = slugify.slugify(basename)
        fname = os.path.join(Run.tmp_artifacts, slug_name + "." + extension)
        figure.savefig(fname, dpi=dpi, format=extension, bbox_inches="tight", **kwargs)

        return Run.add_artifact(
            fname, name=name, step=step, wait_response=wait_response, run_id=run_id, project=project
        )

    @staticmethod
    @online_method
    @functional_alias(alias="add_config")
    def add_config(config: Union[Config, dict[str, Any]], run_id: Optional[str] = None):
        """Attach a config to the run."""
        if isinstance(config, Config):
            config = config.to_dict()
        Run._config = config
        return Run._update(run_id=run_id, config=config)

    @staticmethod
    @online_method
    @functional_alias(alias="add_metric")
    def add_metric(key: str, step: int, value: float):
        """
        Add a metric to the run.

        Args:
            key (str): The key of the metric.
            step (int): The step of the metric.
            value (float): The value of the metric.

        Note:
            The metrics are not pushed immediately. They are pushed when the `Run._push_metrics`
            method is called inside the MetricsManager. See `pyhectiqlab.metrics.MetricsManager`.
        """
        if Run._metrics is None:
            Run._metrics = MetricsManager(push_method=Run._push_metrics, run_id=Run.id)
        return Run._metrics.add(key, step, value)

    @staticmethod
    @online_method
    @functional_alias(alias="get_agg_metrics")
    def get_agg_metrics(key: str, run_id: Optional[str] = None) -> List[Tuple[Union[int, float], Union[int, float]]]:
        """Get the aggregated metrics for the given key."""
        run_id = run_id or Run.id
        if run_id is None:
            return []
        return MetricsManager.get_agg(key, run_id)
    @staticmethod
    @online_method
    def _push_metrics(
        key: str,
        values: List[Tuple[Union[int, float], Union[int, float], Union[int, float]]],
        run_id: Optional[str] = None,
    ):
        """Push the metrics to the server.
        Do not use this method directly. Use the `add_metric` method instead.
        values is a list of tuples (step, value, timestamp).
        """
        run_id = run_id or Run.id
        if key is None or run_id is None or values is None:
            return
        data = dict(name=key, values=values, run=run_id)
        # for value in values:
        #     if len(value) != 3 or any(filter(lambda x: x is None, value)):
        #         logging.error(f"Invalid metric value: {value}")
        #         return
        # print(data)
        Run._client.post("/app/metrics", json=data, wait_response=False)

    @staticmethod
    @online_method
    @functional_alias("start_step")
    def start(
        name: str,
        run_id: Optional[str] = None,
        description: Optional[str] = None,
        metadata: Optional[dict] = None,
        status: Optional[str] = None,
    ):
        """
        Start a new step.

        Args:
            title (str): Title of the step.
        """
        if Step.id is not None:
            Step.end()

        return Step.start(
            name=name,
            run_id=Run.get_id(run_id),
            description=description,
            metadata=metadata,
            status=status,
        )

    @staticmethod
    @online_method
    @functional_alias("end_step")
    def end(run_id: Optional[str] = None, status: Optional[str] = None):
        """End the current step."""
        return Step.end(status=status)

    @staticmethod
    @online_method
    @functional_alias('attach_model')
    def attach_model(
        name: str,
        version: str,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Attach a model to the run.

        Args:
            name (str): The name of the model.
            version (str): The version of the model. If None, the latest version is attached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Model.attach(
            name=name,
            version=version,
            run_id=Run.get_id(run_id),
            project=project,
            wait_response=wait_response,
        )

    @staticmethod
    @online_method
    @functional_alias("detach_model")
    def detach_model(
        name: str,
        version: str,
        run_id: Optional[int] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Detach a model from the run.

        Args:
            name (str): The name of the model.
            version (str): The version of the model. If None, the latest version is detached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """

        return Model.detach(
            name=name, version=version, run_id=Run.get_id(run_id), project=project, wait_response=wait_response
        )

    @staticmethod
    @online_method
    @functional_alias()
    def attach_dataset(
        name: str,
        version: str,
        run_id: Optional[int] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Attach a dataset to the run.

        Args:
            name (str): The name of the dataset.
            version (str): The version of the dataset.  If None, the latest version is detached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Dataset.attach(
            name=name, version=version, run_id=Run.get_id(run_id), project=project, wait_response=wait_response
        )

    @staticmethod
    @online_method
    @functional_alias()
    def detach_dataset(
        name: str,
        version: str,
        run_id: Optional[int] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """
        Detach a dataset to the run.

        Args:
            name (str): The name of the dataset.
            version (str): The version of the dataset.  If None, the latest version is detached. Default: None.
            run_id (str, optional): The ID of the run. If None, the current run is used. Default: None.
            project (str, optional): The project of the run. If None, the current project is used. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Dataset.detach(
            name=name, version=version, run_id=Run.get_id(run_id), project=project, wait_response=wait_response
        )

    @staticmethod
    @online_method
    @functional_alias("set_run_status")
    @clisync.include()
    def set_status(status: str, run_id: Optional[str] = None):
        """Set the run to the given status."""
        Run._update(run_id=run_id, status=status)

    @staticmethod
    @online_method
    @functional_alias("failed")
    @clisync.include()
    def failed(run_id: Optional[str] = None):
        """Set the run to the failed status."""
        Run.set_status(run_id=run_id, status="failed")

    @staticmethod
    @online_method
    @functional_alias("stopped")
    @clisync.include()
    def stopped(run_id: Optional[str] = None):
        """Set the run to the stopped status."""
        Run.set_status(run_id=run_id, status="stopped")

    @staticmethod
    @online_method
    @functional_alias("completed")
    @clisync.include()
    def completed(run_id: Optional[str] = None):
        """Set the run to the completed with success status."""
        Run.set_status(run_id=run_id, status="completed")

    @staticmethod
    @online_method
    @functional_alias("pending")
    @clisync.include()
    def pending(run_id: Optional[str] = None):
        """Set the run to the pending status."""
        Run.set_status(run_id=run_id, status="pending")

    @staticmethod
    @online_method
    @functional_alias("running")
    @clisync.include()
    def running(run_id: Optional[str] = None):
        """Set the run to the running status."""
        Run.set_status(run_id=run_id, status="running")

    @staticmethod
    @online_method
    @functional_alias("training")
    @clisync.include()
    def training(run_id: Optional[str] = None):
        """Set the run to the training status."""
        Run.set_status(run_id=run_id, status="training")

    @staticmethod
    @online_method
    @functional_alias("add_tags_to_run")
    def add_tags(tags: Union[str, List[str]], run_id: Optional[str] = None, project: Optional[str] = None):
        """Add a tag to the run.
        For functional alias, use `attach_tags_to_run`.

        Args:
            title (str): The new title of the run.
            run_id (str): The ID of the run. If None, the current run is used. Default: None.
            project (str): The project of the run. If None, the current project is used. Default: None.
        """
        if isinstance(tags, str):
            tags = [tags]
        Tag.attach_to_run(tags, run_id=Run.get_id(run_id), project=Project.get(project))

    @staticmethod
    @online_method
    @functional_alias("detach_tag_from_run")
    def detach_tag(tag: str, run_id: Optional[str] = None, project: Optional[str] = None):
        """Remove a tag from the run.
        For functional alias, use `detach_tag_from_run`.

        Args:
            title (str): The new title of the run.
            run_id (str): The ID of the run. If None, the current run is used. Default: None.
            project (str): The project of the run. If None, the current project is used. Default: None.
        """
        Tag.detach_from_run(tag=tag, run_id=run_id or Run.id, project=Project.get(project))

    def __str__(self):
        pad = 8
        path = "/" + Project.get() + "/runs/" + str(self.rank)  # TBD
        return (
            f"<Run {self.rank}>"
            f"\n{'Project'.ljust(pad)}: {Project.get()}"
            f"\n{'Block'.ljust(pad)}: {Block.get()}"
            f"\n{'author'.ljust(pad)}: {self._client.auth.user}"
            f"\n{'online'.ljust(pad)}: {self._client.auth.online}"
            f"\n{'url'.ljust(pad)}: {API_URL+path} "
        )



import os
import socket
import asyncio
import logging
from typing import List, Optional

logger = logging.getLogger()


def batched(iterable, n):
    """Yield successive n-sized batches from iterable."""
    for i in range(0, len(iterable), n):
        i1 = min(i + n, len(iterable))
        yield iterable[i:i1]


def list_all_files_in_dir(local_folder: str) -> List[str]:
    """List all files in a directory.
    If the path is a file, the file is returned.

    Args:
        local_folder (str): The folder to list files from.
    """
    if not os.path.exists(local_folder):
        raise FileNotFoundError(f"Directory {local_folder} does not exist.")
    filenames = []
    if not os.path.isdir(local_folder):
        # If the path is a file, return the file
        return [local_folder]

    for el in os.walk(local_folder):
        for f in os.listdir(el[0]):
            complete_path = os.path.join(el[0], f)
            if not os.path.isdir(complete_path):
                if os.path.isfile(complete_path):
                    filenames.append(complete_path)
    return filenames

def extract_host_from_source(source: str) -> str:
    """Extract the resource from a path.

    If the path is a cloud path (e.g., s3://bucket/dataset), the resource is inferred from the path.
    If the path is a local path, the resource is set to "local" if not provided.

    Args:
        path (str): The path to extract the resource from.
        default_resource (str): The resource to compare the path with.
    """
    if source.startswith("s3://"):
        return "s3"
    elif source.startswith("gs://"):
        return "gs"
    else:
        return socket.gethostname()


async def queue_gather(tasks: List[callable], workers: Optional[int] = 10):
    """Gather tasks using a queue and workers.
    The tasks will be placed in a queue and workers will be
    created to process the tasks.
    The workers will be stopped when the queue is empty.

    Example usage:

    ```python
    async def task():
        await asyncio.sleep(1)
        return "done"
    tasks = [task() for _ in range(10)]
    results = await queue_gather(tasks)
    print(results)
    ```

    Args:
        tasks (List[callable]): A list of async methods
        workers (Optional[int], optional): The number of workers. Defaults to 10.
    """

    async def worker(i: int, queue: asyncio.Queue):
        """Worker to process tasks from a queue."""
        results = []
        while queue.empty() == False:
            logger.debug(f"Worker {i}. Queue size: {queue.qsize()}")
            idx, task = await queue.get()
            try:
                result = await task
                results.append((idx, result))
            except Exception as e:
                print(e)
                results.append((idx, None))
            finally:
                queue.task_done()
        return results

    queue = asyncio.Queue()
    [queue.put_nowait((i, task)) for i, task in enumerate(tasks)]
    # Create workers
    workers = [asyncio.create_task(worker(idx, queue)) for idx in range(workers)]
    await queue.join()
    for worker in workers:
        worker.cancel()
    results = await asyncio.gather(*workers, return_exceptions=True)
    flatten_results = [result for worker_results in results for result in worker_results]
    flatten_results.sort(key=lambda x: x[0])
    return [result for (_, result) in flatten_results]



import os
import clisync
import asyncio
import logging
import shutil

from typing import Optional, List, Union

from pyhectiqlab.tag import Tag
from pyhectiqlab.block import Block
from pyhectiqlab.project import Project
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias, online_method
from pyhectiqlab.utils import batched, list_all_files_in_dir
from pyhectiqlab.settings import getenv

logger = logging.getLogger()

class Model:
    _client: Client = Client

    @staticmethod
    @online_method
    @functional_alias("create_model")
    @clisync.include()
    def create(
        name: str,
        path: Optional[str] = None,
        version: Optional[str] = None,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        block: Optional[str] = None,
        description: Optional[str] = None,
        upload: Optional[bool] = True,
    ):
        """Create a model

        Args:
            name (str): Name of the model
            run_id (str, optional):  ID of the run
            project (str, optional): Project of the model
            path (str, optional): Path to the model directory
            description (str, optional): Description of the model
            version (str, optional): Version of the model
            block (str, optional): Block of the model
            upload (bool, optional): Upload the model files. Default: True
        """
        from pyhectiqlab.run import Run

        project = Project.get(project)
        run_id = Run.get_id(run_id)
        block = Block.get(block)

        if project is None:
            logger.error("Project not found, skipping model creation.")
            return

        if name is None:
            raise ValueError("The `name` parameter is required.")

        if block:
            if "::" not in block:
                block = Block.format_id(block, project)

        json = {
            "name": name,
            "description": description,
            "version": version,
            "block": block,
            "project": project,
            "root_run": run_id,
        }

        model = Model._client.post("/app/models", wait_response=True, json=json)
        if not model or "id" not in model:
            return

        if upload and path:
            Model.upload(id=model["id"], path=path)

        Model.attach(name=name, version=version, run_id=run_id, project=project, wait_response=True)
        return model

    @staticmethod
    @online_method
    @functional_alias("upload_model")
    @clisync.include()
    def upload(id: str, path: str):
        """Upload local files to a model hosted in the Lab.

        Args:
            id (str): ID of the model.
            path (str): Path to the model directory.
        """
        all_files = list_all_files_in_dir(path)
        batch_size = 50
        Model._id = model_id = id
        for batch in batched(all_files, batch_size):
            # Create many files
            batch_body = [{"name": os.path.relpath(f, start=path), "num_bytes": os.path.getsize(f)} for f in batch]
            files = Model._client.post(
                f"/app/models/{model_id}/files", wait_response=True, json={"files": batch_body}
            )["results"]

            # For each file in batch, get the policy in files (use name for finding the file)
            sorted_files = []
            for el in batch:
                file = [f.get("upload_policy") for f in files if f["name"] == os.path.relpath(el, start=path)][0]
                sorted_files.append(file)

            # Upload model files
            asyncio.run(Model._client.upload_many(paths=batch, policies=sorted_files))

    @staticmethod
    @online_method
    @functional_alias("retrieve_model")
    @clisync.include()
    def retrieve(
        name: str,
        version: str,
        project: Optional[str] = None,
        fields: Optional[List[str]] = [],
    ):
        """Retrieve a model

        Args:
            name (str): Name of the model
            version (str): Version of the model
            project (str, optional): Project of the model. If None, the project of the current run is used. Default: None.
            fields (dict, optional): Fields to retrieve
            wait_response (bool, optional): Wait for the response from the server
        """
        project = Project.get(project)
        if project is None:
            return

        return Model._client.get(
            "/app/models/retrieve",
            params={
                "name": name,
                "project": project,
                "version": version,
                "fields": fields,
            },
            wait_response=True,
        )

    @staticmethod
    @online_method
    @functional_alias("list_models")
    @clisync.include(wait_response=True)
    def list(
        project: Optional[str] = None,
        search: Optional[str] = None,
        author: Optional[str] = None,
        block: Optional[str] = None,
        blocks: Optional[List[str]] = None,
        keep_latest_version: bool = False,
        fields: Optional[List[str]] = [],
        page: Optional[int] = 1,
        limit: Optional[int] = 100,
        order_by: Optional[str] = "created_at",
        order_direction: Optional[str] = "desc",
        wait_response: bool = False,
    ):
        """List the models

        Args:
            project (str): Project of the model
            block (str, optional): Block of the model
            search (str, optional): Search string
            author (str, optional): Author of the model
            blocks (List[str], optional): Blocks of the model
            keep_latest_version (bool, optional): If True, group by the latest version of model name and return only the latest version of each model name
            fields (List[str], optional): Fields to retrieve
            page (int, optional): Page number
            limit (int, optional): Limit of the models
            order_by (str, optional): Order by
            order_direction (str, optional): Order direction
            wait_response (bool, optional): Wait for the response from the server
        """

        if block is not None:
            if "::" not in block:
                from pyhectiqlab.block import Block

                block = Block.format_id(block, project)

        params = {
            "project": Project.get(project),
            "block": block,
            "search": search,
            "author": author,
            "blocks": blocks,
            "keep_latest_version": keep_latest_version or True,
            "fields": fields,
            "page": page,
            "limit": limit,
            "order_by": order_by,
            "order_direction": order_direction,
        }

        return Model._client.get("/app/models", params=params, wait_response=wait_response)

    @staticmethod
    @online_method
    @functional_alias("update_model")
    @clisync.include(wait_response=True)
    def update(
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        block: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Update a model

        Args:
            id (str): ID of the model
            name (str, optional): Name of the model
            description (str, optional): Description of the model
            version (str, optional): Version of the model
            block (str, optional): Block of the model
            wait_response (bool, optional): Wait for the response from the server
        """
        if not id:
            return
        return Model._client.put(
            f"/app/models/{id}",
            json={"name": name, "description": description, "version": version, "block": block},
            wait_response=wait_response,
        )

    @staticmethod
    @online_method
    @functional_alias("delete_model")
    @clisync.include(wait_response=True)
    def delete(
        id: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Delete a model

        Args:
            id (str): ID of the model
            name (str, optional): Name of the model
            version (str, optional): Version of the model
            project (str, optional): Project of the model
            wait_response (bool, optional): Wait for the response from the server
        """
        project = Project.get(project)
        assert id or (name and version and project), "Either `id` or `name`, `version` and `project` must be provided."
        if not id:
            model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
            if not model:
                return
            id = model["id"]
        return Model._client.delete(f"/app/models/{id}", wait_response=wait_response)

    @staticmethod
    @functional_alias("download_model")
    @clisync.include(wait_response=True)
    def download(
        name: str,
        version: str,
        project: Optional[str] = None,
        path: Optional[str] = None,
        overwrite: bool = False,
        wait_response: bool = True,
    ):
        """Download a model

        Args:
            name (str): Name of the model
            project (str): Project of the model
            version (str): Version of the model
            path (str, optional): Path to download the model
            overwrite (bool, optional): Overwrite the files if they already exist
            wait_response (bool, optional): Wait for the response from the server
        """
        project = Project.get(project)
        if not project:
            logger.error("Project not found, skipping model download.")
            return
        
        path = path or getenv("HECTIQLAB_MODELS_DOWNLOAD", "./")
        path = os.path.join(path, f"{name}-{version}")
        if os.path.exists(path):
            if not overwrite:
                logger.warning(f"Directory {path} already exists. Set `overwrite=True` to overwrite the files.")
                return path

            shutil.rmtree(path)
        os.makedirs(path, exist_ok=True)
        from pyhectiqlab import Run

 

        async def composition():
            model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
            if not model:
                return
            model_id = model["id"]

            if Run.id is not None:
                Model._client.post(f"/app/models/{model_id}/attach-to-run/{Run.id}", wait_response=False)

            num_results = 0
            total_results = 1
            page = 1
            while not num_results == total_results:
                files = Model._client.get(
                    f"/app/models/{model_id}/files/",
                    params={"page": page, "limit": 50, "fields": ["download_url", "name", "num_bytes"]},
                    wait_response=True,
                )
                total_results = files["total_results"]

                await Model._client.download_many(
                    urls=[file["download_url"] for file in files["results"]],
                    local_paths=[os.path.join(path, file["name"]) for file in files["results"]],
                    num_bytes=[file["num_bytes"] for file in files["results"]],
                )
                num_results += len(files["results"])
                page += 1

            Model._client.post(f"/app/models/{model_id}/on-download-completed", wait_response=False)

        res = Model._client.execute(composition, wait_response=wait_response, is_async_method=True)
        if isinstance(res, asyncio.Task):
            return res
        return path

    @staticmethod
    @online_method
    def attach(
        name: str,
        version: str,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = True,
    ):
        """Attach a model to a run

        Args:
            run_id (str): ID of the run.
            name (str): Name of the model.
            version (str): Version of the model.
            project (str, optional): Project of the model. If None, the project of the current run is used. Default: None.
            wait_response (bool, optional): Wait for the response from the server. Default: False.
        """
        from pyhectiqlab.run import Run

        run_id = Run.get_id(run_id)
        if run_id is None:
            return
        project = Project.get(project)
        if not project:
            logger.error("Project not found, skipping model attachment.")
            return

        model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
        if not model:
            return
        model_id = model["id"]
        return Model._client.post(f"/app/models/{model_id}/attach-to-run/{run_id}", wait_response=wait_response)

    @staticmethod
    @online_method
    def detach(
        name: str,
        version: str,
        run_id: str,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Detach a model from a run

        Args:
            name (str): Name of the model.
            version (str): Version of the model.
            run_id (int): ID of the run.
            project (str, optional): Project of the model. If None, the project of the current run is used. Default: None.
            wait_response (bool, optional): Wait for the response from the server. Default: False.
        """
        project = Project.get(project)
        if not project:
            logger.error("Project not found, skipping model detachment.")
            return
        model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
        if not model:
            return
        model_id = model["id"]
        return Model._client.post(f"/app/models/{model_id}/detach-from-run/{run_id}", wait_response=wait_response)

    @staticmethod
    @online_method
    @functional_alias("add_tags_to_model")
    def add_tags(
        tags: Union[str, List[str]],
        name: str,
        version: str,
        project: Optional[str] = None,
    ):
        """Add tags to a model

        Args:
            tags (Union[str, List[str]]): Tags to add.
            name (str): Name of the model.
            version (str): Version of the model.
            project (str): Project of the model. If None, the project of the current run is used. Default: None.
        """
        project = Project.get(project)
        if not project:
            logger.error("Project not found, skipping tag addition.")
            return
        model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
        if not model:
            return
        model_id = model["id"]
        return Tag.attach_to_model(tags=tags, model_id=model_id, project=project)

    @staticmethod
    @online_method
    @functional_alias("detach_tag_from_model")
    def detach_tag(
        tag: str,
        name: str,
        version: str,
        project: Optional[str] = None,
    ):
        """Remove a tag from the model.

        Args:
            title (str): The new title of the run.
            name (str): Name of the model.
            version (str): Version of the model.
            project (str): Project of the model. If None, the project of the current run is used. Default: None.
        """
        project = Project.get(project)
        if not project:
            logger.error("Project not found, skipping tag detachment.")
            return
        model = Model.retrieve(name=name, version=version, project=project, fields=["id"])
        if not model:
            return
        model_id = model["id"]
        Tag.detach_from_model(tag=tag, model_id=model_id, project=project)



import slugify
import clisync
import toml
import os
import logging

from typing import Optional, List
import pyhectiqlab.settings as settings
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import classproperty, online_method, functional_alias

logger = logging.getLogger()


class Project:
    """
    Project class for handling project related operations.

    Args:
        slug (str): Slug of the project.
        configs (str): Path to the config file.
        allow_dirty (bool): Allow dirty state of the project.
        client (Client): Client object.
    """

    _slug: str = settings.getenv("HECTIQLAB_PROJECT", None)
    _configs: str = settings.getenv("HECTIQLAB_CONFIGS", os.path.expanduser("~/.hectiq-lab/configs.toml"))
    _allow_dirty: Optional[bool] = None
    _client: Client = Client

    @classproperty
    def slug(cls) -> str:
        return cls._slug

    @classproperty
    def repo(cls) -> str:
        if not cls._slug:
            return None
        if not os.path.exists(cls._configs):
            return None
        with open(cls._configs, "r") as file:
            config = toml.load(file)
            return config.get(cls._slug, {}).get("repo")

    @classmethod
    def allow_dirty(cls) -> bool:
        if not Client.online():
            return True
        if cls._allow_dirty is None and os.path.exists(cls._configs):
            with open(cls._configs, "r") as file:
                config = toml.load(file)
                cls._allow_dirty = not config.get(cls._slug, {}).get("force_no_git_diff", False)
        return cls._allow_dirty or True

    @staticmethod
    @online_method
    @clisync.include()
    def create(
        name: str,
        repo: Optional[str] = None,
        force_no_git_diff: bool = True,
        write_config: bool = True,
    ) -> Optional[str]:
        """
        Create a project.

        Args:
            name (str): Name of the project.
            repo (str, optional): Repository URL. Default: None.
            force_no_git_diff (bool): Force no git diff. Default: True.
            write_config (bool): Write the config file. Default: True.
        """
        from pyhectiqlab.const import DISABLE_PROJECT_CREATION

        if DISABLE_PROJECT_CREATION:
            logging.error("Project creation is disabled. Continuing anyway...")
            return
        slug = "/".join([slugify.slugify(n) for n in name.split("/")])
        body = {
            "slug": slug,
            "name": name,
            "repo": repo,
            "force_no_git_diff": force_no_git_diff,
        }
        project = Project._client.post("/app/projects", json=body, wait_response=True)
        if project is None:
            logging.error(f"Failed to create project.")
            return
        if write_config:
            Project.write_config(slug)
        logger.info(f"Project `{slug}` created.")
        return slug

    @staticmethod
    @online_method
    def write_config(project: str):
        if not os.path.exists(Project._configs):
            config = {project: {}}
        else:
            config = toml.load(Project._configs)
            config[project] = {}
        project = Project.retrieve(project, fields=["slug", "repo", "force_no_git_diff", "description"])
        config[project["slug"]]["repo"] = project.get("repo")
        config[project["slug"]]["force_no_git_diff"] = project.get("force_no_git_diff")
        config[project["slug"]]["description"] = project.get("description")

        with open(Project._configs, "w") as file:
            toml.dump(config, file)

    @staticmethod
    @online_method
    def exists(slug: str) -> bool:
        project = Project._client.get(f"/app/projects/{slug}", wait_response=True)
        if project is None:
            return False
        return True

    @staticmethod
    @online_method
    @functional_alias("retrieve_project")
    @clisync.include()
    def retrieve(
        slug: str,
        fields: Optional[List[str]] = [],
        update_config: bool = False,
    ) -> Optional[dict]:
        """Retrieve a project info.

        Args:
            slug (str): Slug of the project.
            fields (List): Fields to retrieve.
        """
        info = Project._client.get(f"/app/projects/{slug}", params={"fields": fields}, wait_response=True)
        if update_config:
            Project.write_config(slug)
        return info

    @staticmethod
    @functional_alias("get_project")
    def get(slug: Optional[str] = None):
        return slug or Project._slug

    @staticmethod
    @functional_alias("set_project")
    def set(slug: str):
        if not Project.exists(slug):
            logging.error(f"The project `{slug}` does not exist. Continuing with the current project.")
            Project._client.online(False)
            return
        Project._slug = slug



import httpx
import asyncio
import threading
import logging

import os
import io
from typing import Any, Optional, List, Type, Union
from functools import partial
from pyhectiqlab.auth import Auth
from pyhectiqlab.decorators import classproperty, request_handle
from pyhectiqlab import API_URL
from google.resumable_media.requests import ResumableUpload
import tqdm

logger = logging.getLogger()

ResponseType = Union[dict[str, Any], bytes, Type[None]]



class Client:
    """
    Client singleton for making sync and async requests in the Hectiq Lab API. This client
    can be used in a context manager, or as a singleton. For performing async requests, the object
    must be instanciated.

    Example:
    ```python
    from pyhectiqlab.client import Client
    import asyncio

    client = Client()

    # Sync request
    response = client.get("/app/auth/is-logged-in")

    # Async request
    response = asyncio.run(client.async_get("/app/auth/is-logged-in"))
    ```
    """

    _auth: Auth = None
    _client: httpx.Client = None
    _online: bool = True
    _timeout: float = 30.0  # seconds

    @classproperty
    def auth(cls) -> Auth:
        if cls._auth is None:
            cls._auth = Auth()
        return cls._auth

    @classproperty
    def client(cls) -> httpx.Client:
        if cls._client is None:
            cls._client = httpx.Client(auth=cls.auth, timeout=cls._timeout)
        return cls._client

    @staticmethod
    def is_logged():
        return Client.auth.is_logged()

    @staticmethod
    def online(status: Optional[bool] = None):
        if status is not None:
            Client._online = status
        return Client.auth.online and Client._online

    @staticmethod
    def get(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("get", url, wait_response=wait_response, **kwargs)

    @staticmethod
    def post(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("post", url, wait_response=wait_response, **kwargs)

    @staticmethod
    def patch(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("patch", url, wait_response=wait_response, **kwargs)

    @staticmethod
    def put(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("put", url, wait_response=wait_response, **kwargs)

    @staticmethod
    def delete(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("delete", url, wait_response=wait_response, **kwargs)

    @staticmethod
    async def upload(local_path: str, policy: dict) -> ResponseType:
        """Upload a file."""
        upload_method = policy.get("upload_method")
        if not upload_method:
            return

        async with httpx.AsyncClient(timeout=Client._timeout) as asyncClient:
            if upload_method == "fragment":
                res = await Client.upload_fragment(local_path, policy, client=asyncClient)
            elif upload_method == "single":
                res = await Client.upload_single(
                    local_path, policy.get("policy"), bucket=policy.get("bucket_name"), client=asyncClient
                )
        return res

    @staticmethod
    def get_upload_method(policy: dict) -> dict[str, Any]:
        """Get the upload method from the policy."""
        if "upload_method" in policy:
            return policy.get("upload_method")
        return policy.get("policy", {}).get("upload_method")

    @staticmethod
    async def upload_many(paths: List[str], policies: List[dict]) -> None:
        """Upload many files."""
        single_upload_files = []
        fragment_upload_files = []
        for path, policy in zip(paths, policies):
            if not policy:
                continue
            method = Client.get_upload_method(policy)
            if method == "fragment":
                fragment_upload_files.append((path, policy))
            elif method == "single":
                single_upload_files.append((path, policy))

        async with httpx.AsyncClient() as asyncClient:
            tasks = []
            for path, policy in single_upload_files:
                task = Client.upload_single(
                    path, policy.get("policy"), bucket=policy.get("bucket_name"), client=asyncClient
                )
                tasks.append(asyncio.ensure_future(task))
            for path, policy in fragment_upload_files:
                task = Client.upload_fragment(path, policy, client=asyncClient)
                tasks.append(asyncio.ensure_future(task))
            await asyncio.gather(*tasks)

    @staticmethod
    async def upload_fragment(
        local_path: str, policy: dict, chunk_size: int = 1024 * 1024 * 32, client: Optional[httpx.AsyncClient] = None
    ) -> None:
        """Upload a file in fragments."""
        upload = ResumableUpload(upload_url=policy.get("url"), chunk_size=chunk_size)
        data = open(local_path, "rb").read()
        upload._stream = io.BytesIO(data)
        upload._total_bytes = len(data)
        upload._resumable_url = policy.get("url")

        progress_bar = tqdm.tqdm(total=upload.total_bytes, unit="iB", unit_scale=True)

        bytes_uploaded = 0
        while upload.finished == False:
            method, url, payload, headers = upload._prepare_request()
            if headers.get("content-type") == None:
                headers["content-type"] = "application/octet-stream"
            result = await (client or httpx).request(
                method, url, data=payload, headers=headers, timeout=Client._timeout
            )
            upload._process_resumable_response(result, len(payload))
            progress_bar.update(upload.bytes_uploaded - bytes_uploaded)
            bytes_uploaded = upload.bytes_uploaded
        progress_bar.close()

    @staticmethod
    async def upload_single(
        local_path: str, policy: dict, bucket: str, client: Optional[httpx.AsyncClient] = None
    ) -> ResponseType:
        """Upload a file in a single request."""
        url = policy.get("url")
        content = open(local_path, "rb")
        num_bytes = os.path.getsize(local_path)
        files = {"file": (bucket, content)}
        upload_method = client.post if client else request_handle(httpx.post)
        with tqdm.tqdm(total=num_bytes, unit="iB", unit_scale=True) as pbar:
            res = await upload_method(url, data=policy.get("fields"), files=files)
            pbar.update(num_bytes)
        return res

    @staticmethod
    async def download(
        url: str, local_path: str, num_bytes: Optional[int] = None, client: Optional[httpx.AsyncClient] = None
    ) -> str:
        """Download a file.

        Args:
            url (str): URL of the file to download.
            local_path (str): Local path to save the file (includes the file name)
            num_bytes (Optional[int], optional): Number of bytes to download. Default: None.
        """
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        with open(local_path, "wb") as f:
            with tqdm.tqdm(
                total=num_bytes, unit="iB", unit_scale=True, bar_format="{desc:<5.5}{percentage:3.0f}%|{bar:10}{r_bar}"
            ) as pbar:
                async with (client or httpx.AsyncClient()).stream("GET", url) as r:
                    async for data in r.aiter_bytes():
                        f.write(data)
                        pbar.update(len(data))
        return local_path

    @staticmethod
    async def download_many(urls: List[str], local_paths: List[str], num_bytes: List[int]) -> None:
        """Download many files.

        Args:
            urls (List[str]): URLs of the files to download.
            local_paths (List[str]): Local paths to save the files (includes the file names)
            num_bytes (List[int]): Number of bytes to download.
        """
        async with httpx.AsyncClient() as client:
            tasks = []
            for url, local_path, byt in zip(urls, local_paths, num_bytes):
                task = Client.download(url, local_path, byt, client=client)
                tasks.append(asyncio.ensure_future(task))
            await asyncio.gather(*tasks)

    @staticmethod
    def request(call: str, url: str, wait_response: bool = False, **kwargs) -> ResponseType:
        """Execute a request to the Hectiq Lab API."""
        if not Client.online():
            return

        url = Client.format_url(url)
        method = request_handle(partial(getattr(Client.client, call), url=url))
        return Client.execute(method=method, wait_response=wait_response, **kwargs)

    @staticmethod
    def execute(
        method: callable, wait_response: bool = False, is_async_method: bool = False, **kwargs
    ) -> ResponseType:
        """Execute a request in the background or in the main thread."""
        if wait_response:
            if is_async_method:
                return asyncio.run(method(**kwargs))
            return method(**kwargs)

        def threading_method(**kwargs):
            if is_async_method:
                return asyncio.run(method(**kwargs))
            return method(**kwargs)

        t = threading.Thread(target=threading_method, kwargs=kwargs)
        t.start()
        return

    @staticmethod
    def format_url(url: str) -> str:
        if not url.startswith("http") and url[0] == "/":
            url = API_URL + url
        return url



from pyhectiqlab.artifact import Artifact
from .utils import expose_static_methods

"""
Functional run API. To expose a function, use the `exposed_method` decorator in the `Run` singleton.
"""

locals().update(expose_static_methods(Artifact))



import sys
import click
from .auth import auth_group
from .configure import config_group
from clisync import CliSync


def main():
    import pyhectiqlab
    from pyhectiqlab import Run, Model, Dataset
    from pyhectiqlab.project import Project
    from pyhectiqlab.artifact import Artifact
    from pyhectiqlab.tag import Tag
    from pyhectiqlab.versions import PackageVersion

    # enable block and project creation for the cli
    pyhectiqlab.const.DISABLE_BLOCK_CREATION = pyhectiqlab.const.DISABLE_PROJECT_CREATION = False

    group = CliSync(module=pyhectiqlab, classes=[Run, Model, Dataset, Artifact, Tag, Project, PackageVersion])
    cli = click.CommandCollection(
        sources=[auth_group, config_group, group], help="👋 Hectiq Lab CLI. Documentation at https://docs.hectiq.ai."
    )
    # Standalone mode is False so that the errors can be caught by the runs
    cli(standalone_mode=False)
    sys.exit()


if __name__ == "__main__":
    main()



import os
import click

from typing import Optional


@click.group()
def config_group():
    pass


@config_group.command("info", help="Display information about your current project")
def display_info():
    import toml
    from pyhectiqlab.auth import is_authenticated
    from pyhectiqlab.settings import getenv

    if not is_authenticated():
        click.secho("You need to authenticate first.", fg="red")
        return

    configs_path = getenv("HECTIQLAB_CONFIGS", os.path.expanduser("~/.hectiq-lab/configs.toml"))
    with open(configs_path, "r") as f:
        configs = toml.load(f)

    click.secho("🔧 Configuration for your projects.")
    for project, config in configs.items():
        click.secho(f"[{project}]", fg="green")
        for k, v in config.items():
            click.secho(f"  - {k}: {v}", fg="blue")


@config_group.command("configure", help="Configure the current project")
@click.option("--project", help="Project name", default=None)
@click.option("--repos", help="Repos to be used, separated by commas.", default=None)
@click.option("--model-path", help="Path where the model are saved.", default=None)
@click.option("--dataset-path", help="Path where the dataset are saved.", default=None)
@click.option("--configs-path", help="Path where the configs are saved.")
@click.option("--allow-dirty", help="Whether to allow dirty repos.", is_flag=True)
def configure(
    project: Optional[str],
    repos: Optional[str],
    model_path: Optional[str],
    dataset_path: Optional[str],
    configs_path: Optional[str],
    allow_dirty: Optional[bool],
):
    import toml

    from pyhectiqlab.auth import is_authenticated
    from pyhectiqlab.settings import getenv

    click.secho("🔧 Configuring your project.")
    if not is_authenticated():
        click.secho("You need to authenticate first.", fg="red")
        return

    info = {}
    info["project"] = click.prompt("Project name", type=str) if project is None else project
    if info["project"] is None:
        click.secho("Project name is required.", fg="red")
        return

    info["allow_dirty"] = (
        click.prompt("Allow dirty repos", type=bool, show_default=True) if allow_dirty is None else allow_dirty
    )

    info["repos"] = (
        click.prompt("Repos to be used, separated by commas", type=str, default="", show_default=True)
        if repos is None
        else repos
    ) or None
    if info["repos"] is not None:
        info["repos"] = info["repos"].split(",")

    info["model_path"] = (
        click.prompt("Path to the model", type=str, show_default=True, default="")
        if model_path is None
        else model_path
    ) or None
    if info["model_path"] is not None:
        info["model_path"] = os.path.abspath(info["model_path"])

    info["dataset_path"] = (
        click.prompt("Path to the dataset", type=str, show_default=True, default="")
        if dataset_path is None
        else dataset_path
    ) or None
    if info["dataset_path"] is not None:
        info["dataset_path"] = os.path.abspath(info["dataset_path"])

    # Writing to the configs file
    configs_path = configs_path or getenv("HECTIQLAB_CONFIGS", os.path.expanduser("~/.hectiq-lab/configs.toml"))
    with open(configs_path, "r") as f:
        configs = toml.load(f)
    info = {k: v for k, v in info.items() if v is not None}
    if info["project"] not in configs:
        configs[info["project"]] = {}

    for k, v in info.items():
        if v is None:
            info.pop(k)
        if k == "project":
            continue
        if k in configs[info["project"]] and configs[info["project"]][k] == v:
            continue
        elif k in configs[info["project"]]:
            click.secho(f"- {k}={configs[info['project']][k]}.", fg="red")
        click.secho(f"+ {k}={v}.", fg="green")
        configs[info["project"]][k] = v

    with open(configs_path, "w") as f:
        toml.dump(configs, f)
    click.secho("✅ Configuration saved.", fg="green")
    conda_env = getenv("CONDA_PREFIX_2")

    if conda_env is None or not click.prompt(
        f"Do you want to sync the configuration to your conda environment ({conda_env})?",
        type=bool,
        default=True,
        show_choices=True,
    ):
        return

    # Syncing the configuration to the conda environment
    env_sh = os.path.join(conda_env, "etc/conda/activate.d/env_vars.sh")
    with open(env_sh, "r") as f:
        lines = f.readlines()

    def find(lines, prefix):
        val = list(filter(lambda x: x.startswith(prefix), lines))
        if len(val) == 0:
            return None
        return val[0]

    begin = lines.index("# Hectiq Lab configuration\n") if "# Hectiq Lab configuration\n" in lines else len(lines)
    end = lines.index("# End configuration\n") if "# End configuration\n" in lines else len(lines)

    config_lines = ["# Hectiq Lab configuration\n"]
    config_lines.append(f"export HECTIQLAB_PROJECT='{info['project']}'\n")
    if "allow_dirty" in info:
        config_lines.append(f"export HECTIQLAB_ALLOW_DIRTY={int(info['allow_dirty'])}\n")

    if "repos" in info:
        config_lines.append(f"export HECTIQLAB_REPOS='{','.join(info['repos'])}'\n")
    if "model_path" in info:
        config_lines.append(f"export HECTIQLAB_MODELS_DOWNLOAD='{info['model_path']}'\n")
    if "dataset_path" in info:
        config_lines.append(f"export HECTIQLAB_DATASETS_DOWNLOAD='{info['dataset_path']}'\n")
    config_lines.append("# End configuration\n")

    with open(env_sh, "w") as f:
        f.writelines(lines[:begin] + config_lines + lines[end:])
    click.secho("✅ Configuration synced to conda environment.", fg="green")



__version__ = "3.0.11"

from pyhectiqlab.settings import getenv

API_URL = getenv("HECTIQLAB_API_URL", "https://api.lab.hectiq.ai")

from pyhectiqlab.config import Config
from pyhectiqlab.dataset import Dataset
from pyhectiqlab.model import Model
from pyhectiqlab.run import Run
from pyhectiqlab.step import Step
from pyhectiqlab.artifact import Artifact
from pyhectiqlab.project import Project
from pyhectiqlab import functional

from pyhectiqlab.logging import setup_logging

setup_logging()


def debug_mode():
    from pyhectiqlab.client import Client

    Client.online(False)


__all__ = [
    "API_URLs",
    "Artifact",
    "Dataset",
    "Step",
    "Model",
    "Run",
    "Config",
    "Block",
    "Project",
    "functional",
    "debug_mode",
]



import pytest
import os

import logging
logger = logging.getLogger()
from utils import random_uuid

from pyhectiqlab.project import Project

Project.set("hectiq-ai/test")
logger.info("🚨 Testing pyhectiqlab.")
logger.info(f"Version: {__import__('pyhectiqlab').__version__}")
logger.info(f"Path: {__import__('pyhectiqlab').__file__}")
logger.info(f"Project: {Project.get()}")

@pytest.fixture
def source():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "dummy/"))

@pytest.fixture
def save_path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "data"))

@pytest.fixture
def clear_data(data_path):
    yield
    import shutil
    shutil.rmtree(data_path, ignore_errors=True)

def test_dataset_create():
    from pyhectiqlab import Dataset
    dataset = Dataset.create(name="test", source="dummy")
    logger.info(f"Dataset created: {dataset}")

def test_dataset_create_cli():
    import os
    project = Project.get()
    os.system(f"hectiq-lab Dataset.create --name test --source dummy --project {project}")

def test_dataset_create_with_run():
    from pyhectiqlab import Dataset, Run
    run = Run(title=f"test {random_uuid()}")
    dataset = Dataset.create(name="test", source="dummy")
    logger.info(f"Dataset created with run: {dataset}")

def test_dataset_create_and_upload(source):
    from pyhectiqlab import Dataset
    dataset = Dataset.create(name="test", source=source, project="hectiq-ai/test", upload=True)
    logger.info(f"Dataset created and uploaded: {dataset}")

def test_dataset_create_and_upload_cli(source):
    import os
    project = Project.get()
    os.system(f"hectiq-lab Dataset.create --name test --source {source} --project {project} --upload")

def test_download(save_path):
    from pyhectiqlab import Dataset
    dataset = Dataset.download(name="test", 
                               version="1.2", 
                               project="hectiq-ai/test", 
                               path=save_path, 
                               wait_response=True)
    logger.info(f"Dataset downloaded: {dataset}")

def test_download_with_run(save_path):
    from pyhectiqlab import Dataset, Run
    run = Run(title=f"test {random_uuid()}")
    from pyhectiqlab.run import Run
    run_id = Run.get_id()
    logger.info(f"Run ID: {run_id}")
    dataset = Dataset.download(name="test", 
                               version="1.2", 
                               project="hectiq-ai/test", 
                               path=save_path, 
                               wait_response=True)
    logger.info(f"Dataset downloaded with run: {dataset}")

def test_download_cli(save_path):
    import os
    project = Project.get()
    os.system(f"hectiq-lab Dataset.download --name test --version 1.2 --project {project} --path {save_path} ")

def test_dataset_retrieve():
    from pyhectiqlab import Dataset
    dataset = Dataset.retrieve(name="test-dataset", version="1.3", project="hectiq-ai/test")
    assert dataset["name"] == "test-dataset"
    assert dataset["version"] == "1.3"


def test_dataset_add_tag():
    from pyhectiqlab import Dataset
    Dataset.add_tags(tags=["test", "dataset"], name="test-dataset", version="1.3", project="hectiq-ai/test")


if __name__ == "__main__":
    pass



import pytest
from utils import random_uuid
from mock_client import mock_client


@pytest.fixture
def slug():
    return "hectiq-ai/test2"


@pytest.fixture
def teardown_config():
    yield
    import os

    path = os.path.abspath(os.path.join(os.path.dirname(__file__), "dummy/configs.toml"))
    if os.path.exists(path):
        os.remove(path)


def test_project_create():
    from pyhectiqlab import Project
    from pyhectiqlab.client import Client
    import uuid

    id = str(uuid.uuid4())[:6]
    with mock_client(Project) as p:
        p.create(name="tests/" + id, write_config=False)


def test_project_write_config(slug, teardown_config):
    import os
    import toml
    from pyhectiqlab import Project

    current = Project._configs
    config_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "dummy/configs.toml"))

    Project._configs = config_path
    Project.write_config(slug)
    config = toml.load(open(Project._configs, "r"))
    assert slug in config
    Project._configs = current


def test_project_retrieve(slug):
    from pyhectiqlab import Project

    project = Project.retrieve(slug)
    assert project["name"] == slug
    assert project["slug"] == slug


def test_project_exists(slug):
    from pyhectiqlab import Project
    import uuid

    fake_id = str(uuid.uuid4())[:6]

    assert Project.exists(slug)
    assert not Project.exists(slug + fake_id)


def test_project_set_get(slug):
    from pyhectiqlab import Project

    assert Project.get() == None
    Project.set(slug)
    assert Project.get() == slug


if __name__ == "__main__":
    pass



import os
import clisync
import asyncio
import shutil
import logging

from typing import Optional, List, Union

from pyhectiqlab.tag import Tag
from pyhectiqlab.project import Project
from pyhectiqlab.block import Block
from pyhectiqlab.block import Block
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias, online_method
from pyhectiqlab.settings import getenv
from .utils import batched, list_all_files_in_dir, extract_host_from_source


class Dataset:
    _client: Client = Client

    @staticmethod
    @online_method
    @functional_alias("create_dataset")
    @clisync.include()
    def create(
        name: str,
        source: str,
        host: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        block: Optional[str] = None,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        upload: Optional[bool] = False,
    ):
        """Create a dataset.

        Supported cloud storages:
        - Amazon S3
        - Google Cloud Storage

        Args:
            name (str): Name of the dataset.
            source (str): Path to the dataset. If your dataset is located in a local directory, the source should be the directory path (e.g., "/path/to/dataset"). If
                the dataset is located in a cloud storage, the source should be the URL of the dataset (e.g., "s3://bucket/dataset").
            host (str, optional): Host of the dataset. Default: None. If the dataset is located in a local directory,
                the host could be your hostname or leave it empty. If the dataset is located in a cloud storage, the host should be the cloud storage name ("s3" or "gs")
            description (str, optional): Description of the dataset. Default: None.
            version (str]): Version of the dataset. Default: None.
            block (str, optional): Block of the dataset. Default: None.
            run_id (str, optional): Run of the dataset. Default: None.
            project (str, optional): Project of the dataset. Default: None.
            upload (bool): If True, upload the dataset to the Lab. Default: False.
        """
        from pyhectiqlab.run import Run

        project = Project.get(project)
        run_id = Run.get_id(run_id)
        block = Block.get(block)
        if project is None:
            logging.error("Project not found, skipping dataset creation.")
            return

        if name is None:
            raise ValueError("The `name` parameter is required.")

        if source is None:
            raise ValueError("The `source` parameter is required.")
        source = os.path.abspath(source)
        host = extract_host_from_source(source)

        if block:
            if "::" not in block:
                block = Block.format_id(block, project)

        data = {
            "name": name,
            "description": description,
            "version": version,
            "host": host,
            "source": source,
            "block": block,
            "project": project,
            "root_run": run_id,
        }
        dataset = Dataset._client.post("/app/datasets", wait_response=True, json=data)
        if upload and host not in ["s3", "gs"]:
            Dataset.upload(id=dataset["id"], source=source)

        Dataset.attach(name=name, version=version, run_id=run_id, project=project)
        return dataset

    @staticmethod
    @online_method
    @functional_alias("upload_dataset")
    @clisync.include()
    def upload(id: str, source: str):
        """Upload local files to a dataset hosted in the Lab.

        Args:
            id (str): ID of the dataset.
            path (str): Path to the directory containing the files to upload.
        """
        all_files = list_all_files_in_dir(source)
        batch_size = 50
        Dataset._id = dataset_id = id
        for batch in batched(all_files, batch_size):
            # Create many files
            batch_body = [{"name": os.path.relpath(f, start=source), "num_bytes": os.path.getsize(f)} for f in batch]
            files = Dataset._client.post(
                f"/app/datasets/{dataset_id}/files", wait_response=True, json={"files": batch_body}
            )["results"]

            # For each file in batch, get the policy in files (use name for finding the file)
            sorted_files = []
            for el in batch:
                file = [f.get("upload_policy") for f in files if f["name"] == os.path.relpath(el, start=source)][0]
                sorted_files.append(file)
            # Upload dataset files
            asyncio.run(Dataset._client.upload_many(paths=batch, policies=sorted_files))

    @staticmethod
    @online_method
    @functional_alias("retrieve_dataset")
    @clisync.include()
    def retrieve(name: str, version: str, project: str, fields: Optional[List[str]] = None):
        """Retrieve a dataset

        Args:
            name (str): Name of the dataset
            project (str): Project of the dataset
            version (str): Version of the dataset
            fields (list[str], optional): Fields to retrieve. Default: None.
        """
        body = Dataset._client.get(
            "/app/datasets/retrieve",
            params={
                "name": name,
                "project": Project.get(project),
                "version": version,
                "fields": fields or [],
            },
            wait_response=True,
        )
        return body

    @staticmethod
    @online_method
    @functional_alias("list_datasets")
    @clisync.include()
    def list(
        project: Optional[str] = None,
        search: Optional[str] = None,
        author: Optional[str] = None,
        block: Optional[str] = None,
        blocks: Optional[List[str]] = None,
        keep_latest_version: bool = False,
        fields: Optional[List[str]] = [],
        page: Optional[int] = 1,
        limit: Optional[int] = 100,
        order_by: Optional[str] = "created_at",
        order_direction: Optional[str] = "desc",
    ):
        """List the datasets

        Args:
            project (str, optional): Project of the dataset.
            block (str, optional): Block of the dataset.
            search (str, optional): Search string.
            author (str, optional): Author of the dataset.
            blocks (list[str], optional): Blocks of the dataset.
            keep_latest_version (bool, optional): If True, group by the latest version of dataset name and return only the latest version of each dataset name.
            fields (list[str], optional): Fields to retrieve.
            page (int, optional): Page number.
            limit (int, optional): Limit of the datasets.
            order_by (str, optional): Order by.
            order_direction (str, optional): Order direction.
            wait_response (bool): Wait for the response from the server. Default: False.
        """
        block = Block.get(block)
        if block is not None:
            if "::" not in block:
                block = Block.format_id(block, project)

        params = {
            "project": project,
            "block": block,
            "search": search,
            "author": author,
            "blocks": blocks,
            "keep_latest_version": keep_latest_version or True,
            "fields": fields,
            "page": page,
            "limit": limit,
            "order_by": order_by,
            "order_direction": order_direction,
        }

        return Dataset._client.get("/app/datasets", params=params, wait_response=True)

    @staticmethod
    @online_method
    @functional_alias("update_dataset")
    @clisync.include(wait_response=True)
    def update(
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        block: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Update a dataset

        Args:
            id (str): ID of the dataset.
            name (str, optional): Name of the dataset.
            description (str, optional): Description of the dataset.
            version (str, optional): Version of the dataset.
            block (str, optional): Block of the dataset.
            wait_response (bool): Wait for the response from the server. Default: False
        """
        if not id:
            return
        return Dataset._client.put(
            f"/app/datasets/{id}",
            json={"name": name, "description": description, "version": version, "block": block},
            wait_response=wait_response,
        )

    @staticmethod
    @online_method
    @functional_alias("delete_dataset")
    @clisync.include(wait_response=True)
    def delete(
        id: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Delete a dataset.

        Args:
            id (str): ID of the dataset.
            name (str, optional): Name of the dataset.
            version (str, optional): Version of the dataset.
            project (str, optional): Project of the dataset.
            wait_response (bool): Wait for the response from the server. Default: False
        """
        assert id or (name and version and project), "Either `id` or `name`, `version` and `project` must be provided."
        if not id:
            dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
            if not dataset:
                return
            id = dataset["id"]
        return Dataset._client.delete(f"/app/datasets/{id}", wait_response=wait_response)

    @staticmethod
    @online_method
    @functional_alias("download_dataset")
    @clisync.include(wait_response=True)
    def download(
        name: str,
        version: str,
        project: Optional[str] = None,
        path: Optional[str] = None,
        overwrite: bool = False,
        wait_response: bool = True,
    ):
        """Download a dataset

        Args:
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            project (str, optional): Project of the dataset.
            path (str): Path to download the dataset. Default: "./".
            overwrite (bool): Overwrite the existing files. Default: False.
            wait_response (bool): Wait for the response from the server. Default: False.
        """
        from pyhectiqlab import Run

        path = path or getenv("HECTIQLAB_MODELS_DOWNLOAD", "./")
        path = os.path.join(path, f"{name}-{version}")
        if os.path.exists(path):
            if not overwrite:
                logging.warning(f"Directory {path} already exists. Set `overwrite=True` to overwrite the files.")
                return path

            shutil.rmtree(path)

        os.makedirs(path, exist_ok=True)
        run_id = Run.get_id()
        project = Project.get(project)
        if run_id is not None:
            Dataset.attach(name=name, version=version, run_id=run_id, project=project)

        async def composition():
            dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
            if not dataset:
                return
            dataset_id = dataset["id"]

            num_results = 0
            total_results = 1
            page = 1
            while not num_results == total_results:
                files = Dataset._client.get(
                    f"/app/datasets/{dataset_id}/files/",
                    params={"page": page, "limit": 50, "fields": ["download_url", "name", "num_bytes"]},
                    wait_response=True,
                )
                total_results = files["total_results"]

                await Dataset._client.download_many(
                    urls=[file["download_url"] for file in files["results"]],
                    local_paths=[os.path.join(path, file["name"]) for file in files["results"]],
                    num_bytes=[file["num_bytes"] for file in files["results"]],
                )
                num_results += len(files["results"])
                page += 1

            Dataset._client.post(f"/app/datasets/{dataset_id}/on-download-completed", wait_response=False)

        Dataset._client.execute(composition, wait_response=wait_response, is_async_method=True)
        return path

    @staticmethod
    @online_method
    def attach(
        name: str,
        version: str,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Attach a dataset to a run

        Args:
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            run_id (str): ID of the run.
            project (str, optional): Project of the dataset.
            wait_response (bool): Wait for the response from the server. Default: False.
        """
        from pyhectiqlab.run import Run

        run_id = Run.get_id(run_id)
        if run_id is None:
            return
        dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
        if dataset is None:
            return
        return Dataset._client.post(
            f"/app/datasets/{dataset['id']}/attach-to-run/{run_id}", wait_response=wait_response
        )

    @staticmethod
    @online_method
    def detach(
        name: str,
        version: str,
        run_id: str,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Detach a dataset from a run

        Args:
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            run_id (str): ID of the run.
            project (str, optional): Project of the dataset.
            wait_response (bool, optional): Wait for the response from the server. Default: False.
        """
        if not run_id:
            return
        dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
        if not dataset:
            return
        dataset_id = dataset["id"]
        return Dataset._client.post(
            f"/app/datasets/{dataset_id}/detach-from-run/{run_id}", wait_response=wait_response
        )

    @staticmethod
    @online_method
    @functional_alias("add_tags_to_dataset")
    def add_tags(
        tags: Union[str, List[str]],
        name: str,
        version: str,
        project: Optional[str] = None,
        wait_response: Optional[bool] = False,
    ):
        """Add tags to a dataset

        Args:
            tags (Union[str, List[str]]): Tags to add.
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            project (str, optional): Project name. Default: None.
            wait_response (bool, optional): Wait for the response from the server. Default: False.
        """
        project = Project.get(project)
        dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
        if not dataset:
            return
        dataset_id = dataset["id"]
        return Tag.attach_to_dataset(tags=tags, dataset_id=dataset_id, project=project, wait_response=wait_response)

    @staticmethod
    @online_method
    @functional_alias("detach_tag_from_dataset")
    def detach_tag(
        tag: str,
        name: str,
        version: str,
        project: Optional[str] = None,
        wait_response: Optional[bool] = False,
    ):
        """Remove a tag from the run.
        For functional alias, use `detach_tag_from_run`.

        Args:
            title (str): The new title of the run.
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            project (str, optional): Project name. Default: None.
        """
        project = Project.get(project)
        dataset = Dataset.retrieve(name=name, version=version, project=project, fields=["id"])
        if not dataset:
            return
        dataset_id = dataset["id"]
        Tag.detach_from_dataset(tag=tag, dataset_id=dataset_id, project=project, wait_response=wait_response)
